import torch

from deepproblog.dataset import DataLoader
from deepproblog.engines import ExactEngine
from deepproblog.evaluate import get_confusion_matrix
from deepproblog.examples.Coins.data.dataset import train_dataset, test_dataset
from deepproblog.model import Model
from deepproblog.network import Network
from deepproblog.train import train_model
from deepproblog.utils.standard_networks import smallnet
from deepproblog.utils.stop_condition import Threshold, StopOnPlateau

batch_size = 5
loader = DataLoader(train_dataset, batch_size)
lr = 1e-4
coin_network1 = smallnet(num_classes=2, pretrained=True)
coin_net1 = Network(coin_network1, "net1", batching=True)
coin_net1.optimizer = torch.optim.Adam(coin_network1.parameters(), lr=lr)
coin_network2 = smallnet(num_classes=2, pretrained=True)
coin_net2 = Network(coin_network2, "net2", batching=True)
coin_net2.optimizer = torch.optim.Adam(coin_network2.parameters(), lr=lr)

model = Model("model.pl", [coin_net1, coin_net2])
model.add_tensor_source("train", train_dataset)
model.add_tensor_source("test", test_dataset)
model.set_engine(ExactEngine(model), cache=True)
train_obj = train_model(
    model,
    loader,
    StopOnPlateau("Accuracy", warm_up=10, patience=10)
    | Threshold("Accuracy", 1.0, duration=2),
    log_iter=100 // batch_size,
    test_iter=100 // batch_size,
    test=lambda x: [("Accuracy", get_confusion_matrix(x, test_dataset).accuracy())],
    infoloss=0.25,
)



import re

for train in [2, 3, 4]:
    for test in [8, 64]:
        for mode in ["train", "test", "dev"]:
            with open("train{}_test{}/{}.txt".format(train, test, mode)) as f:
                text = f.read()
            matches = re.findall("\[([0-9 ]*)\]\t\[([0-9 ]*) (\d*)\]", text)
            text = list()
            for match in matches:
                l_sorted = [int(i) for i in match[0].split(" ")]
                l = [int(i) for i in match[1].split(" ")]
                l.reverse()
                l_sorted.reverse()
                text.append("forth_sort({},{}).".format(l, l_sorted))

            with open("train{}_test{}_{}.txt".format(train, test, mode), "w") as f:
                f.write("\n".join(text))



from __future__ import annotations

import re
from collections import defaultdict
from time import time
from typing import Any, Dict, List, Optional, Tuple, Union, Generator, Iterable

from problog.core import ProbLogObject
from problog.logic import And
from problog.logic import Term, Clause, Or, AnnotatedDisjunction, Var
from problog.logic import term2list, ArithmeticError
from problog.program import LogicProgram
from pyswip import Prolog, registerForeign, Variable

from .heuristics import Heuristic
from .swip import parse, pyswip_to_term

ALL_FACTS = "fa(_,_,_,_)"

ALL_CLAUSES = "cl(_,_,_)"

ids = 0
current_program = None


def get_heuristic_func(heuristic_obj):
    def get_heuristic(node, heuristic: Variable):
        node = pyswip_to_term(node)
        value = heuristic_obj.get_value(node)
        heuristic.unify(value)

    get_heuristic.arity = 2
    return get_heuristic


class SWIProgramException(Exception):
    """Exception for prolog issues from SWIProgram"""


# Regex for query variable replacement
_RE_QUERY = re.compile(r"\$VAR\((.*?)\)")


class SWIProgram(ProbLogObject):
    FactOrClause = Union[Tuple[str, int, str, str], Tuple[str, int, float, str, str]]

    def __init__(
        self,
        db: Optional[LogicProgram] = None,
        heuristic: Optional[Heuristic] = None,
        parent: Optional[SWIProgram] = None,
    ):
        global ids
        self.id = "p{}".format(ids)
        self.groups = 0
        ids += 1
        self.prolog = Prolog()
        if parent is not None:
            self.facts_and_clauses = parent.facts_and_clauses
            self.db = parent.db
            self.ad_heads = parent.ad_heads
            self.i = parent.i
            self.d = parent.d
        else:

            self.facts_and_clauses: List[Optional[SWIProgram.FactOrClause]] = []
            self.db = db
            self.ad_heads = defaultdict(list)
            self.i = 0
            self.d = dict()
            self.parse_db()
        registerForeign(get_heuristic_func(heuristic), name="get_heuristic_extern")
        self.heuristic = heuristic

    # ASSERTING

    def assert_clause(self, clause):
        self.prolog.assertz(self.to_prolog(clause))

    def assert_fact(self, fact):
        self.prolog.assertz(self.to_prolog(fact))

    def add_program(
        self, db: Iterable[Union[Term, Or, Clause]]
    ) -> Generator[Tuple[str, int], None, None]:
        """
        Add all clauses and facts from a LogicProgram
        :return: The indices of the added facts and clauses.
        """
        for n in db:
            if type(n) is Clause:
                yield self.add_clause(n)
            elif type(n) is Or:
                yield from self.add_or(n)
            elif type(n) is Term:
                yield self.add_fact(n)
            elif type(n) is AnnotatedDisjunction:
                yield from self.add_or(n)
            else:
                raise SWIProgramException("Unhandled node type ", type(n))

    def add_fact(
        self,
        node: Term,
        group_id: Optional[int] = None,
        ad_i: Optional[int] = None,
        ad_vars: Optional[List[Var]] = None,
    ):
        probability = node.probability
        if probability is None:
            probability = 1.0
        ad_identifier = "none"
        if group_id is not None:
            ad_vars = [str(x) for x in ad_vars]
            ad_identifier = "ad({},{},[{}])".format(group_id, ad_i, ",".join(ad_vars))

        return self._add_fact(
            probability, str(node.with_probability(None)), ad_identifier
        )

    def _add_fact(
        self, probability: float, fact, ad_identifier, i: Optional[int] = None
    ) -> FactOrClause:
        if i is None:
            i = self.new_entry()
            self.facts_and_clauses.append(("fa", i, probability, fact, ad_identifier))
        else:
            assert self.facts_and_clauses[i - 1] is None
            self.facts_and_clauses[i - 1] = ("fa", i, probability, fact, ad_identifier)
        global current_program
        new_fact = ("fa", i, probability, fact, ad_identifier)
        if current_program == self:
            self.assert_fact(new_fact)
        return new_fact

    def add_or(self, node: Union[AnnotatedDisjunction, Or]):
        if type(node) is AnnotatedDisjunction:
            if node.body.functor == "true":
                disjuncts = node.heads
            else:
                raise SWIProgramException("ADs with bodies not yet supported")
        else:
            disjuncts = node.to_list()
        group_id = self.groups
        self.groups += 1
        variables = set()
        for disjunct in disjuncts:
            variables |= disjunct.variables()
        variables = list(variables)
        for i, disjunct in enumerate(disjuncts):
            yield self.add_fact(disjunct, group_id, i, variables)

    def add_clause(self, node: Clause, i=None) -> Optional[FactOrClause]:
        if node.head.functor == "_directive":
            return None
        if type(node.body) is Term:
            body = [node.body]
        elif type(node.body) is And:
            body = node.body.to_list()
        else:
            raise SWIProgramException("Unhandled body type")
        return self._add_clause(
            str(node.head), "[{}]".format(",".join(str(x) for x in body)), i=i
        )

    def _add_clause(
        self, head: str, body: str, i: Optional[int] = None
    ) -> FactOrClause:
        global current_program
        if i is None:
            i = self.new_entry()
            self.facts_and_clauses.append(("cl", i, head, body))
        else:
            assert self.facts_and_clauses[i - 1] is None
            self.facts_and_clauses[i - 1] = ("cl", i, head, body)
        new_clause = ("cl", i, head, body)
        if current_program == self:
            self.assert_clause(new_clause)
        return new_clause

    def assert_all(self):
        global current_program

        if current_program != self:
            current_program = self
            self.retract_all()
            for line in self.get_lines():
                self.prolog.assertz(line)

    # RETRACTING

    def retract_fact(self, i: int):
        self.prolog.retractall("fa({},_,_,_)".format(i))

    def remove_fact(self, i: int):
        global current_program
        if current_program == self:
            self.retract_fact(i)
        return self._remove_clause_or_fact(i)

    def _remove_clause_or_fact(self, i: int):
        removed = self.facts_and_clauses[i - 1]
        self.facts_and_clauses[i - 1] = None
        return removed

    def remove_clause(self, i: int):
        global current_program
        if current_program == self:
            self.retract_clause(i)
        return self._remove_clause_or_fact(i)

    def retract_clause(self, i: int):
        self.prolog.retractall("cl({},_,_)".format(i))

    def retract_all(self):
        self.prolog.retractall(ALL_CLAUSES)
        self.prolog.retractall(ALL_FACTS)

    # Proof building

    def build_tree(self, proof, target):
        name, proof = proof.args
        try:
            key = target.names[name]
        except KeyError:
            key = target.add_or([], placeholder=True, readonly=False)
            target.names[name] = key
            target.add_name(name, key)
        t = proof.functor
        if t == "::":
            i = (proof.args[0], proof.args[2])
            try:
                p = float(proof.args[1])
            except ArithmeticError:
                p = proof.args[1]
            group = proof.args[3]
            if group.functor == "ad":
                group = group.args[0], tuple(term2list(group.args[2], deep=False))
            else:
                group = None
            new_name = Term("fact", *i)
            try:
                new = target.names[new_name]
            except KeyError:
                i = target.get_next_atom_identifier()
                new = target.add_atom(i, p, group=group)
                target.names[new_name] = new
                target.add_name(new_name, new)
            target.add_disjunct(key, new)
        elif t == "and":
            body = proof.args[0]
            body = term2list(body, False)
            new_name = Term("and", *(p.args[0] for p in body))
            try:
                new = target.names[new_name]
                for b in body:
                    self.build_tree(b, target)
            except KeyError:
                new = target.add_and([self.build_tree(b, target) for b in body])
                target.names[new_name] = new
                target.add_name(new_name, new)
            target.add_disjunct(key, new)
        elif t == "foreign":
            new = target.TRUE
            target.add_disjunct(key, new)
        elif t == "neg":
            new = -self.build_tree(proof.args[0], target)
            target.add_disjunct(key, new)
        elif t == "cycle":
            pass
        elif t == "builtin":
            new = target.TRUE
            target.add_disjunct(key, new)
        elif t == "extern":
            new = target.TRUE
            target.add_disjunct(key, new)
        elif t == "nn":
            # net, inp, i = proof.args
            target.add_atom()
        else:
            raise SWIProgramException("Unhandled node type " + str(proof))
        return key

    def add_proof_trees(self, trees, target, label=None):
        target.names = dict()
        for tree in trees:
            depth, tree = tree.args
            query, proof = tree.args
            key = self.build_tree(tree, target)
            target.add_name(query, key, label=label)
        return target

    def query(self, query: str, profile=0) -> List[Dict[str, Any]]:
        """Perform a Prolog query and translate result back to Problog objects

        :param query: Query to perform
        :param profile: Allows enabling profiling.

        :return: A list of results to the query.
        """
        self.assert_all()

        # Replaces $VAR(X) with actual variables
        # Needed when specified queries are non ground
        query = _RE_QUERY.sub(r"X\1", query)
        #
        start = 0
        if profile > 0:
            start = time()
            if profile > 1:
                # query = 'profile((between(1,100,_),{},fail);true)'.format(query)
                query = f"profile({query})"
        result = list(self.prolog.query(query))
        if profile > 0:
            print(f"Query: {query} answered in {time() - start} seconds")
            if profile > 1:
                input()
        if result:
            out = []
            for r in result:
                out_partial = {}
                for k in r:
                    v = result[0][k]
                    if type(v) is list:
                        if len(v) > 1 and isinstance(v[0], str):
                            raise TypeError("Oops, it appears you are using the wrong version of PySwip.\n"
                                            "Please make sure you are using PySwip from https://github.com/ML-KULeuven/pyswip\n"
                                            "To install, first remove your current PySwip version, then install the correct version.\n"
                                            "You can try doing\n'pip install git+https://github.com/ML-KULeuven/pyswip'.\n"
                                            "For some reason, this does not always resolve the issue. If you still get this error,\n"
                                            "clone the repo locally, then compile from your local codebase doing\n"
                                            "'pip install [path to local clone]'.")
                        out_partial[k] = [p for p in term2list(parse(result[0][k]))]
                    else:
                        out_partial[k] = parse(v)
                out.append(out_partial)
            return out
        else:
            return []

    # Other

    @staticmethod
    def to_prolog(f: FactOrClause) -> str:
        if f[0] == "fa":
            return "fa({},{},{},{})".format(*f[1:])
        elif f[0] == "cl":
            return "cl({},{},{})".format(*f[1:])

    def get_lines(self) -> List[str]:
        lines = [self.to_prolog(c) for c in self.facts_and_clauses if c is not None]
        return lines

    def __str__(self):
        return "\n".join(line + "." for line in self.get_lines())

    def parse_db(self):
        """
        Parse the database (LogicProgram) into a valid SWI-Prolog
        :return: List of added clauses (Also updates the current object)
        """
        if self.db is not None:
            return list(self.add_program(self.db))

    @staticmethod
    def registerForeign(func, name, arity=None, **kwargs):
        registerForeign(func, name, arity, **kwargs)

    def extend(self):
        print("Extending")
        return SWIProgram(parent=self)

    def new_entry(self) -> int:
        self.i += 1
        return self.i



from random import choices

import torch
import torch.nn as nn
from typing import List
from problog.logic import term2list


def tokenize2(text, entities, vocab):
    text = text.strip('"').split(" ")
    tokenized = []
    for i, word in enumerate(text):
        try:
            word = int(word)
            try:
                word = entities.index(word) + 1
            except ValueError:
                word = 0

            tokenized.append(word)

        except ValueError:
            try:
                tokenized.append(vocab[word])
            except KeyError:
                tokenized.append(vocab["OOV"])
    return tokenized


def tokenize(text, entity, vocab):
    text = text.strip('"').split(" ")
    tokenized = []
    indices = []
    for i, word in enumerate(text):
        try:
            word = int(word)
            if word == entity:
                word = 1
                indices.append(i)
            else:
                word = 0

            tokenized.append(word)

        except ValueError:
            try:
                tokenized.append(vocab[word])
            except KeyError:
                tokenized.append(vocab["OOV"])
    return tokenized, indices


def tokenize_cloze(text, entities, vocab, nr_entities=10):
    text = text.strip('"').split(" ")
    entity_tokens = ["ENT{}".format(x) for x in range(nr_entities)]
    entity_tokens = choices(entity_tokens, k=len(entities))
    tokenized = []
    indices: List[List[int]] = [[] for _ in range(len(entities))]
    for i, word in enumerate(text):
        try:
            word = int(word)
            indices[entities.index(word)].append(i)
            word = vocab[entity_tokens[entities.index(word)]]
            tokenized.append(word)
        except ValueError:
            try:
                tokenized.append(vocab[word])
            except KeyError:
                tokenized.append(vocab["OOV"])
    return tokenized, indices


class Encoder(nn.Module):
    def __init__(self, vocab, hidden_size, embed_size=None, p_drop=0.0, weights=None):
        super(Encoder, self).__init__()
        self.hidden_size = hidden_size
        if embed_size is None:
            if weights is None:
                self.embed_size = hidden_size
            else:
                self.embed_size = weights.shape[1]
        else:
            self.embed_size = embed_size
        self.vocab = vocab
        self.lstm = nn.GRU(
            self.embed_size,
            hidden_size,
            bidirectional=True,
            batch_first=True,
            num_layers=2,
            dropout=p_drop,
        )
        self.dropout = nn.Dropout(p_drop)
        if weights is None:
            self.embedding = nn.Embedding(len(self.vocab), self.embed_size)
        else:
            self.embedding = nn.Embedding.from_pretrained(weights)

    def forward(self, text, ent1, ent2):
        text = text.value.strip('"')
        ent1, ent2 = int(ent1), int(ent2)
        x, indices = tokenize_cloze(text, [ent1, ent2], self.vocab)
        x = self.embedding(torch.LongTensor(x))
        x, _ = self.lstm(x.unsqueeze(0))
        x = x.view(-1, 2, self.hidden_size)
        forward, backward = x[:, 0, :], x[:, 1, :]
        out = []
        for j, e in enumerate([ent1, ent2]):
            i = indices[j]
            if i:
                i = torch.LongTensor(i)
                embedding = torch.cat([forward[i, :], backward[-i - 1, :]])
                out.append(torch.mean(embedding, 0))
            else:
                out.append(None)
        return out


class RelNet(nn.Module):
    def __init__(self, in_size, mid_size, out_size=11, activation=True):
        super(RelNet, self).__init__()
        self.in_size = in_size
        self.mid_size = mid_size
        self.out_size = in_size if out_size is None else out_size
        self.embed = nn.Sequential(nn.Linear(2 * self.in_size, self.out_size), )
        self.activation_layer = nn.Softmax(dim=-1)
        self.activation = activation

    def forward(self, ex, ey):
        x = torch.cat([ex, ey], 0)
        x = self.embed(x.unsqueeze(0))
        if self.activation:
            x = self.activation_layer(x)
        return x[0]


class GenderNet(nn.Module):
    def __init__(self, vocab, hidden_size, embed_size=None):
        super(GenderNet, self).__init__()
        self.hidden_size = hidden_size
        self.embed_size = embed_size
        if embed_size is None:
            self.embed_size = hidden_size

        self.vocab = vocab
        self.lstm = nn.GRU(
            self.embed_size,
            hidden_size,
            bidirectional=False,
            batch_first=True,
            num_layers=1,
        )
        self.embedding = nn.Embedding(len(self.vocab), self.embed_size)
        self.classification = nn.Sequential(
            nn.Linear(self.hidden_size, 2), nn.Softmax(-1)
        )

    def forward(self, text, ent):
        text = [x.args[1] for x in term2list(text, False)]
        text = " ".join(t.value.strip('"') + " ." for t in text)
        ent = int(ent)
        x, indices = tokenize_cloze(text, [ent], self.vocab)
        x = self.embedding(torch.LongTensor(x))
        x, _ = self.lstm(x.unsqueeze(0))
        x = x[:, -1, :]
        x = self.classification(x)
        return x[0]



import re

for filename in ["train", "test", "dev"]:
    out_lines = []
    with open(filename + ".txt") as f:
        for line in f:
            solution, question = line.rstrip().split("\t")
            solution = int(float(solution))
            numbers = [int(i) for i in re.findall(" (\d+)", question)]
            out_lines.append(
                'wap("{}",{},{},{},{}).'.format(question, *numbers, solution)
            )
    with open("{}.pl".format(filename), "w") as f:
        f.write("\n".join(out_lines))



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from deepproblog.train import TrainObject


class StopCondition(object):
    def is_stop(self, train_object: "TrainObject"):
        raise NotImplementedError()

    def __add__(self, other: "StopCondition"):
        return Or(self, other)

    def __or__(self, other: "StopCondition"):
        return Or(self, other)

    def __mul__(self, other: "StopCondition"):
        return And(self, other)

    def __and__(self, other: "StopCondition"):
        return And(self, other)


class EpochStop(StopCondition):
    def __init__(self, max_epoch: int):
        self.max_epoch = max_epoch

    def __str__(self):
        return "for {} epoch(s)".format(self.max_epoch)

    def is_stop(self, train_object: "TrainObject"):
        return train_object.epoch >= self.max_epoch


class StopOnPlateau(StopCondition):
    def __init__(self, attribute: str, delta=0.01, patience=5, warm_up=5):
        self.attribute = attribute
        self.delta = delta
        self.patience = patience
        self.best = warm_up
        self.warm_up = warm_up

    def __str__(self):
        return "until plateau in {}".format(self.attribute)

    def is_stop(self, train_object: "TrainObject"):
        data = train_object.logger.get_attribute(self.attribute)
        if len(data) <= self.warm_up:
            return False
        if self.best == -1 or (
            data[-1] > data[self.best] + data[self.best] * self.delta
        ):
            self.best = len(data) - 1

        if len(data) - self.best >= self.patience:
            print("No improvement for {} steps. Stopping.".format(self.patience))
            return True
        return False


class StopOnNoChange(StopCondition):
    def __init__(self, attribute: str, delta=0.01, patience=5, warm_up=5):
        self.attribute = attribute
        self.delta = delta
        self.patience = patience
        self.last_change = warm_up
        self.warm_up = warm_up

    def __str__(self):
        return "until no change in {}".format(self.attribute)

    def is_stop(self, train_object: "TrainObject"):
        data = train_object.logger.get_attribute(self.attribute)
        if len(data) <= self.warm_up:
            return False
        if abs(data[-1] - data[self.last_change]) > data[self.last_change] * self.delta:
            self.last_change = len(data) - 1

        if len(data) - self.last_change > self.patience:
            print("No change for {} steps. Stopping.".format(self.patience))
            return True
        return False


class Or(StopCondition):
    def __init__(self, *criteria: "StopCondition"):
        self.criteria = criteria

    def __str__(self):
        return " or ".join([str(c) for c in self.criteria])

    def is_stop(self, train_object: "TrainObject"):
        for c in self.criteria:
            if c.is_stop(train_object):
                return True
        return False


class And(StopCondition):
    def __init__(self, *criteria: "StopCondition"):
        self.criteria = criteria

    def __str__(self):
        return " and ".join([str(c) for c in self.criteria])

    def is_stop(self, train_object: "TrainObject"):
        for c in self.criteria:
            if not c.is_stop(train_object):
                return False
        return True


class Threshold(StopCondition):
    def __init__(self, attribute: str, maximum, duration=1):
        self.attribute = attribute
        self.max = maximum
        self.no_data = 0
        self.duration = duration

    def __str__(self):
        return "until {} reaches {}".format(self.attribute, self.max)

    def is_stop(self, train_object: "TrainObject"):
        data = train_object.logger.get_attribute(self.attribute)
        if len(data) == 0:
            self.no_data += 1
            if self.no_data > 2:
                pass
                # print('Received no data about {} for {} steps'.format(self.attribute, self.no_data))
            return False
        for i in range(self.duration):
            if data[-i - 1] < self.max:
                return False
        return True



#! /usr/bin/env python

from setuptools import setup, find_packages
from pathlib import Path

readme = (Path(__file__).parent / 'README.md').read_text()

setup(
    name="deepproblog",
    version="2.0.6",
    description="DeepProbLog: Problog with neural networks",
    long_description=readme,
    long_description_content_type='text/markdown',
    url="https://github.com/ML-KULeuven/deepproblog",
    author="DeepProbLog team",
    author_email="robin.manhaeve@cs.kuleuven.be",
    license="Apache Software License",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Prolog",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="prolog probabilistic logic neural-symbolic problog deepproblog",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    zip_safe=False,
    include_package_data=True,
    install_requires=["pysdd",
                      "problog",
                      "torch",
                      "torchvision"],
    extras_require={
#        "approximate": ["pyswip @ git+https://github.com/ML-KULeuven/pyswip.git#egg=pyswip"],
        "examples": ["Pillow"],
        "tests": ["pytest"],
    },
)



import random
from abc import ABC, abstractmethod
from math import ceil
from pathlib import Path
from typing import List, Optional, Tuple, TextIO, Union, Callable, Iterable

import problog
from PIL import Image
from problog.logic import Term, Constant

from deepproblog.query import Query


class Dataset(ABC):
    __slots__ = ()

    def __str__(self):
        return "\n".join(str(self.to_query(i)) for i in range(min(len(self), 5)))

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def to_query(self, i: int) -> Query:
        """

        :param i: Index i
        :return: The i-th example as a Query object
        """
        pass

    def to_queries(self) -> List[Query]:
        """

        :return: A list of all queries in the dataset.
        """
        return [self.to_query(i) for i in range(len(self))]

    def subset(self, i: int, j: Optional[int] = None) -> "Dataset":
        """

        :param i: index i
        :param j: index j
        :return: If j is None, returns a subset with the indices [0,i], else returns a subset with the indices [i, j]
        """
        if j is None:
            j = i
            i = 0
        return Subset(self, i, j)

    def __add__(self, other: "Dataset") -> "Dataset":
        """

        :param other: The other dataset.
        :return: Returns a dataset that is the combination of self and other
        """
        return Extension(self, other)

    def fold(self, n: int, i: int) -> Tuple["Dataset", "Dataset"]:
        """

        :param n: The number of folds to make.
        :param i: Which of the folds is the held-out set.
        :return: A tuple of the training fold and test fold datasets.
        """
        return Fold(self, n, i, False), Fold(self, n, i, True)

    def to_file_repr(self, i: int) -> str:
        """

        :param i: index
        :return: Returns a string representation of file i suitable for writing to a file. Defaults to to_query.
        """
        return str(self.to_query(i).to_text_query())

    def write_to_file(self, f: TextIO):
        """

        :param f: File to write the dataset. to
        """
        for i in range(len(self)):
            f.write(self.to_file_repr(i) + "\n")

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __getitem__(self, index):
        pass


class DataLoader(object):
    __slots__ = ("dataset", "batch_size", "length", "shuffle", "epoch", "rng", "i")

    def __init__(
        self, dataset: Dataset, batch_size: int, shuffle: bool = True, seed=None
    ):
        """

        :param dataset: The dataset that this loader will iterate over.
        :param batch_size: The batch size.
        :param shuffle: If true, the queries are shuffled, otherwise, they returned in order.
        :param seed: Seed for random shuffle.
        """
        self.dataset = dataset.to_queries()
        self.batch_size = batch_size
        self.length = len(self.dataset)
        self.shuffle = shuffle
        self.dataset = self.dataset[: self.length]
        self.epoch = 0
        self.rng = random.Random(seed)
        self._set_iter()

    def _shuffling(self):
        if callable(self.shuffle):
            return self.shuffle(self.epoch)
        return self.shuffle

    def _set_iter(self):
        if self._shuffling():
            indices = list(range(self.length))
            self.rng.shuffle(indices)
            self.i = iter(indices)
        else:
            self.i = iter(range(self.length))

    def __next__(self):
        if self.i is None:
            self.epoch += 1
            self._set_iter()
            raise StopIteration
        batch = list()
        try:
            for i in range(self.batch_size):
                batch.append(self.dataset[next(self.i)])
            return batch
        except StopIteration:
            if len(batch) == 0:
                self.epoch += 1
                self._set_iter()
                raise StopIteration
            else:
                self.i = None
            return batch

    def __iter__(self):
        return self

    def __len__(self):
        return int(ceil(self.length / self.batch_size))

    def __repr__(self):
        return "DataLoader: " + str(self.dataset[0])


class Subset(Dataset):
    def __getattr__(self, item):
        return getattr(self.dataset, item)

    def to_query(self, i, *args):
        return self.dataset.to_query(i + self.i, *args)

    def to_file_repr(self, i):
        return self.dataset.to_file_repr(i + self.i)

    def __getitem__(self, item):
        # if item >= len(self):
        #     raise IndexError()
        return self.dataset[item + self.i]

    def __len__(self):
        return self.j - self.i

    def __init__(self, dataset, i, j):
        self.i = i
        self.j = min(j, len(dataset))
        self.dataset = dataset


class Extension(Dataset):
    def __len__(self):
        return len(self.a) + len(self.b)

    def __init__(self, a, b):
        self.a = a
        self.b = b

    def __getitem__(self, item):
        if item < len(self.a):
            return self.a[item]
        else:
            return self.b[item - len(self.a)]

    def to_query(self, i, *args):
        if i < len(self.a):
            return self.a.to_query(i, *args)
        else:
            return self.b.to_query(i - len(self.a), *args)

    def to_file_repr(self, i):
        if i < len(self.a):
            return self.a.to_file_repr(i)
        else:
            return self.b.to_file_repr(i - len(self.a))


class Fold(Dataset):
    def __len__(self):
        return self._len

    def to_file_repr(self, i):
        return self.parent.to_file_repr(self._get_index(i))

    def _get_index(self, i):
        i, j = i // (len(self.indices)), i % (len(self.indices))
        return i * self.n + self.indices[j]

    def to_query(self, i):
        return self.parent.to_query(self._get_index(i))

    def __init__(self, parent, n, i, split):
        self.parent = parent
        self.n = n
        self.i = i
        if split:
            self.indices = [i]
        else:
            self.indices = [x for x in range(n) if x != i]

        self._parent_len = len(parent)
        self._len = len(parent) // n * len(self.indices)
        extra = len(parent) % n
        if i < extra:
            extra = 1 if split else extra - 1
        else:
            extra = 0 if split else extra

        self._len += extra
        self.split = split


class MutatingDataset(Dataset):
    """
    Generic dataset adapter that mutates an underlying dataset.

    Intended use cases involve generating noisy datasets as well as negative examples.
    """

    __slots__ = ("inner_dataset", "mutator")

    def __init__(self, inner_dataset: Dataset, mutator: Callable[[int, Query], Query]):
        super().__init__()
        self.inner_dataset = inner_dataset
        self.mutator = mutator

    def __len__(self):
        return len(self.inner_dataset)

    def to_query(self, i: int) -> Query:
        return self.mutator(i, self.inner_dataset.to_query(i))

    def __getitem__(self, item):
        raise NotImplementedError("__getitem__ is not implemented for MutatingDataset")


class NoiseMutatorDecorator:
    """Dataset mutator that will mutate with a certain probability"""

    __slots__ = ("p", "seed", "inner_mutator")

    def __init__(
        self,
        p: float,
        inner_mutator: Callable[[int, Query], Query],
        seed: Optional[int] = None,
    ):
        """Constructor

        :param p: Probability with which to mutate the sample
        :param inner_mutator: Function that does actual mutation.
        :param seed: Seed for RNG
        """
        self.p = p
        if seed is None:
            seed = random.SystemRandom().randint(0, 2**64)
        self.seed = seed
        self.inner_mutator = inner_mutator

    def __call__(self, index: int, query: Query) -> Query:
        rng = random.Random(self.seed ^ index)
        if rng.random() < self.p:
            return self.inner_mutator(index, query)
        else:
            return query


class NumericIncorrectOutputMutator:
    """Dataset mutator that replaces numeric output with an incorrect value. Only constants are supported."""

    __slots__ = ("domain", "change_p", "seed")

    def __init__(self, domain: List[int], change_p=False, seed: Optional[int] = None):
        """Constructor

        :param domain: Domain of outputs to choose from
        :param change_p: If true, set the new query to have p=1-original.p
        :param seed: Random seed
        """
        self.domain = domain
        self.change_p = change_p
        if seed is None:
            seed = random.SystemRandom().randint(0, 2**64)
        self.seed = seed

    def __call__(self, index: int, query: Query) -> Query:
        # We want a stable behaviour given a specific index
        rng = random.Random(self.seed ^ index)
        new_args = list(query.query.args)
        for i in query.output_ind:
            arg = new_args[i]
            assert isinstance(arg, Constant)
            v = arg.value
            new_v = v
            while new_v == v:
                new_v = rng.choice(self.domain)
            new_args[i] = Constant(new_v)

        new_term = Term(query.query.functor, *new_args)
        return Query(
            query=new_term,
            substitution=query.substitution,
            p=(1.0 - query.p) if self.change_p else query.p,
            output_ind=query.output_ind,
        )


class ImageDataset(Dataset, ABC):
    def __init__(self, root, extension="png", transform=None):
        super().__init__()
        self.root = Path(root)
        self.transform = transform
        self.extension = extension

    def __getitem__(self, index):
        if type(index) is tuple:
            index = index[0]
        p = self.root / "{}.{}".format(index, self.extension)
        with open(p, "rb") as f:
            img = Image.open(f)
            img = img.convert("RGB")
            if self.transform is not None:
                img = self.transform(img)
        return img


def load(filename: Union[Path, str], max_size=None):
    parser = problog.parser.PrologParser(problog.program.ExtendedPrologFactory())
    all_queries = list()
    with open(filename) as f:
        for i, line in enumerate(f):
            if max_size is not None and i >= max_size:
                break
            parsed: Iterable[Term] = parser.parseString(line)
            for term in parsed:
                query_prob = 1.0
                if term.probability:
                    query_prob = term.probability.value
                    term = term.with_probability()
                if term.functor == "subs":
                    t = term.args[0]
                    subs = problog.logic.term2list(term.args[1], False)
                    sub_dict = dict()
                    for sub in subs:
                        a, b = problog.logic.term2list(sub, False)
                        sub_dict[a] = b
                    all_queries.append(Query(t, substitution=sub_dict, p=query_prob))
                else:
                    all_queries.append(Query(term, p=query_prob))
    return all_queries


class QueryDataset(Dataset):
    """Dataset that loads queries from a file or a list"""

    def __len__(self):
        return len(self.queries)

    def to_query(self, i):
        return self.queries[i]

    def __init__(self, queries: Union[List[Query], Path, str]):
        super().__init__()
        if isinstance(queries, list):
            self.queries = queries
        else:
            self.queries = load(queries)



from typing import List, Union

import torch
from numpy.random import choice
from problog.engine import DefaultEngine
from problog.extern import problog_export, problog_export_nondet
from problog.logic import (
    Term,
    AnnotatedDisjunction,
    term2list,
    Clause,
    And,
    Or,
    Var,
    Constant,
    list2term,
)
from problog.program import SimpleProgram

from deepproblog.engines.engine import Engine
from deepproblog.network import Network

# EXTERN = '{}_extern_'
EXTERN = "{}_extern_nocache_"


def wrap_tensor(x, store, name: Union[Term, List[Term]] = None):
    if type(x) is list:
        if name is None:
            return list2term([wrap_tensor(e, store) for e in x])
        else:
            return list2term([wrap_tensor(x[i], store, name[i]) for i in range(len(x))])
    else:
        if name is None:
            return Term("tensor", Constant(store.store(x)))
        else:
            return Term("tensor", store.store(x, name))


def create_with_substitution(formula, second, translation, key):
    if key in translation:
        return translation[key]
    node = second.get_node(key)
    t = type(node).__name__
    if t == "conj":
        return formula.add_and(
            [
                create_with_substitution(formula, second, translation, c)
                for c in node.children
            ],
            name=node.name,
        )
    elif t == "disj":
        return formula.add_or(
            [
                create_with_substitution(formula, second, translation, c)
                for c in node.children
            ],
            name=node.name,
        )
    else:
        raise (Exception("Unknown node ", node))


def get_predicate(net):
    def predicate(inputs):
        domain = net.domain
        output = net([term2list(inputs, False)])[0]
        net.cache[inputs] = output
        if net.eval_mode:
            _, result = torch.topk(output, net.k, 0)
            result = [domain[int(r)] for r in result]
        else:
            result = choice(
                domain, min(net.k, len(domain)), False, output.detach().numpy()
            ).tolist()
        return result

    return predicate


def get_det_predicate(net: Network, engine: Engine):
    def det_predicate(arguments):
        output = net([term2list(arguments, False)])[0]
        tensor_name = Term("nn", Term(net.name), arguments)
        # FIXME named tensor outputs for lists
        return wrap_tensor(output, engine.tensor_store, name=tensor_name)

    return det_predicate


class ExactEngine(Engine):
    def __init__(self, model):
        Engine.__init__(self, model)
        self.engine = DefaultEngine()

    def prepare(self, db):
        translated = SimpleProgram()
        for e in db:
            new_es = [e]
            if type(e) is Term or type(e) is Clause:
                p = e.probability
                if p is not None and p.functor == "nn":
                    if len(p.args) == 4:
                        new_es = self.create_nn_predicate_ad(e)
                    elif len(p.args) == 3:
                        new_es = self.create_nn_predicate_det(e)
                    elif len(p.args) == 2:
                        new_es = self.create_nn_predicate_fact(e)
                    else:
                        raise ValueError(
                            "A neural predicate with {} arguments is not supported.".format(
                                len(p.args)
                            )
                        )
            for new_e in new_es:
                translated.add_clause(new_e)
        translated.add_clause(
            Clause(
                Term("_directive"),
                Term("use_module", Term("library", Term("lists.pl"))),
            )
        )
        clause_db = self.engine.prepare(translated)
        problog_export.database = clause_db
        for network in self.model.networks:
            if self.model.networks[network].det:
                signature = ["+term", "-term"]
                func = get_det_predicate(self.model.networks[network], self)
                problog_export(*signature)(
                    func, funcname=EXTERN.format(network), modname=None
                )
            elif self.model.networks[network].k is not None:
                signature = ["+term", "-list"]
                problog_export(*signature)(
                    get_predicate(self.model.networks[network]),
                    funcname="{}_extern_nocache_".format(network),
                    modname=None,
                )

        return clause_db

    def ground(self, query, label=None, repeat=1, **kwargs):
        db = self.model.solver.program
        if not self.model.solver.cache.cache:
            query = query.substitute()
        ground = self.engine.ground(db, query.query, label=label)
        return ground

    def create_nn_predicate_ad(self, e):
        p = e.probability
        net, inputs, output, domain = p.args
        network = self.model.networks[str(net)]
        network.domain = term2list(domain, False)
        new_terms = []
        heads = []
        for j, domain_element in enumerate(term2list(domain, False)):
            head = Term(
                e.functor if network.k is None else e.functor + "_AD",
                *e.args,
                p=p.with_args(net, inputs, Constant(j))
            )
            head = head.apply_term({output: domain_element})
            heads.append(head)

        if type(e) is Clause:
            ad = AnnotatedDisjunction(heads, e.body)

        else:
            ad = Or.from_list(heads)

        new_terms.append(ad)

        if network.k is not None:
            body = []
            head = e.with_probability(None)
            if type(e) is Clause:
                body.append(e.body)
            body.append(Term(EXTERN.format(net), inputs, Var("Selected")))
            body.append(Term("member", output, Var("Selected")))
            body.append(Term(e.functor + "_AD", *e.args))
            new_e = Clause(head, And.from_list(body))
            new_terms.append(new_e)
        return new_terms

    def create_nn_predicate_fact(self, e):
        p = e.probability
        net, inputs = p.args
        network = self.model.networks[str(net)]
        return [e]

    def create_nn_predicate_det(self, e):
        p = e.probability
        net, inputs, output = p.args
        network = self.model.networks[str(net)]
        network.det = True
        if network.k is not None:
            raise ValueError(
                "k should be None for deterministc network {}".format(str(net))
            )
        head = e.with_probability(None)
        body = Term(EXTERN.format(net), inputs, output)
        return [Clause(head, body)]

    def register_foreign(self, func, function_name, arity_in, arity_out):
        signature = ["+term"] * arity_in + ["-term"] * arity_out
        problog_export.database = self.model.solver.program
        problog_export(*signature)(func, funcname=function_name, modname=None)

    def register_foreign_nondet(self, func, function_name, arity_in, arity_out):
        signature = ["+term"] * arity_in + ["-term"] * arity_out
        problog_export.database = self.model.solver.program
        problog_export_nondet(*signature)(func, funcname=function_name, modname=None)

    def get_hyperparameters(self) -> dict:
        return {"type": "ExactEngine"}

    @staticmethod
    def can_cache() -> bool:
        return True



import pytest
from deepproblog.utils.standard_networks import DummyNet, DummyTensorNet
from problog.logic import Term, Var, Constant
from deepproblog.engines import ExactEngine, ApproximateEngine
from deepproblog.model import Model
from deepproblog.query import Query
from deepproblog.network import Network

import torch

program = """
nn(dummy1,[X],Y,[a,b,c]) :: net1(X,Y).
nn(dummy2,[X]) :: net2(X).
nn(dummy3,[X],Y) :: net3(X,Y).
nn(dummy4,[X,Y]) :: net4(X,Y).

test1(X1,Y1,X2,Y2) :- net1(X1,Y1), net1(X2,Y2).
test2(X1,X2) :- net2(X1), net2(X2).
test3(X,Y) :- net3(X,Y).
"""

dummy_values1 = {Term("i1"): [0.8, 0.15, 0.05], Term("i2"): [0.2, 0.3, 0.5]}
dummy_net1 = Network(DummyNet(dummy_values1), "dummy1")

dummy_values2 = {Term("i1"): [0.8], Term("i2"): [0.4]}
dummy_net2 = Network(DummyNet(dummy_values2), "dummy2")

dummy_values3 = {Term("i1"): [1.0, 2.0, 3.0, 4.0], Term("i2"): [-1.0, 0.0, 1.0]}
dummy_net3 = Network(DummyNet(dummy_values3), "dummy3")


dummy_tensors = {(Term("a"),): torch.Tensor([0.1, 0.2, 0.3, 0.4]), (Term("b"),): torch.Tensor([0.25, 0.25, 0.25, 0.25])}


class IndexNet(torch.nn.Module):

    def forward(self, t, index):
        # index = int(index)
        index = torch.LongTensor([int(i) for i in index])
        return t.index_select(dim=1, index=index)


dummy_net4 = Network(IndexNet(), "dummy4", batching=True)


@pytest.fixture(
    params=[
        {
            "name": "approximate",
            "engine_factory": lambda model: ApproximateEngine(
                model, 10, ApproximateEngine.geometric_mean
            ),
            "cache": False,
        },
        {"name": "no_cache", "engine_factory": lambda model: ExactEngine(model), "cache": False},
        {"name": "cache", "engine_factory": lambda model: ExactEngine(model), "cache": True},
    ]
)
def model(request) -> Model:
    """Simple fixture creating both the approximate and the exact engine"""
    if ApproximateEngine is None and request.param["name"] == "approximate":
        pytest.skip("ApproximateEngine is not available as PySWIP is not installed")
    model = Model(program, [dummy_net1, dummy_net2, dummy_net3, dummy_net4], load=False)
    engine = request.param["engine_factory"](model)
    model.set_engine(engine, cache=request.param["cache"])
    model.add_tensor_source('dummy', dummy_tensors)
    return model


def test_model_basics(model):
    # These should be set up after running the fixture.
    assert model.solver is not None
    assert model.program is not None


def test_ad_network(model: Model):
    q1 = Query(Term("test1", Term("i1"), Term("a"), Term("i2"), Term("b")))  # p = 0.24
    q2 = Query(Term("test1", Term("i1"), Term("a"), Term("i2"), Term("a")))  # p = 0.16
    q3 = Query(Term("test1", Term("i1"), Term("a"), Term("i1"), Term("b")))  # p = 0
    results = model.solve([q1, q2, q3])
    r1 = float(results[0].result[q1.query])
    r2 = float(results[1].result[q2.query])
    r3 = float(results[2].result[q3.query])
    assert pytest.approx(0.24) == r1
    assert pytest.approx(0.16) == r2
    assert pytest.approx(0) == r3


def test_fact_network(model: Model):
    q1 = Query(Term("test2", Term("i1"), Term("i2")))  # p = 0.32
    q2 = Query(Term("test2", Term("i1"), Term("i1")))  # p = 0.8
    results = model.solve([q1, q2])
    r1 = float(results[0].result[q1.query])
    r2 = float(results[1].result[q2.query])
    assert pytest.approx(0.32) == r1
    assert pytest.approx(0.8) == r2


def test_det_network(model: Model):
    results = model.solve([Query(Term("test3", Term("i1"), Var("X"))), Query(Term("test3", Term("i2"), Var("X")))])
    r1 = list(results[0].result)[0].args[1]
    r2 = list(results[1].result)[0].args[1]
    r1 = model.get_tensor(r1)
    r2 = model.get_tensor(r2)
    assert all(r1.detach().numpy() == [1.0, 2.0, 3.0, 4.0])
    assert all(r2.detach().numpy() == [-1.0, 0.0, 1.0])


def test_det_network_substitution(model: Model):
    if not model.solver.cache.cache:
        q1 = Query(Term("test3", Term("a"), Var("X")), {Term("a"): Term("i1")})
        q2 = Query(Term("test3", Term("a"), Var("X")), {Term("a"): Term("i2")})
        results = model.solve([q1, q2])
        r1 = list(results[0].result)[0].args[1]
        r2 = list(results[1].result)[0].args[1]
        r1 = model.get_tensor(r1)
        r2 = model.get_tensor(r2)
        assert all(r1.detach().numpy() == [1.0, 2.0, 3.0, 4.0])
        assert all(r2.detach().numpy() == [-1.0, 0.0, 1.0])

def test_multi_input_network(model: Model):
    dummy_tensor = lambda x: Term("tensor", Term("dummy", x))
    q1 = Query(Term("net4", dummy_tensor(Term("a")), Constant(1)))
    q2 = Query(Term("net4", dummy_tensor(Term("b")), Constant(2)))
    results = model.solve([q1, q2])
    r1 = float(results[0].result[q1.query])
    r2 = float(results[1].result[q2.query])
    assert pytest.approx(0.2) == r1
    assert pytest.approx(0.25) == r2


import math
from typing import Optional

import torch

from deepproblog.optimizer import Optimizer
from deepproblog.semiring import Semiring, Result
from problog.logic import Constant, Term


def get_hook(optimizer: Optimizer, i):
    def hook(grad):
        optimizer.add_parameter_gradient(i, grad)

    return hook


class GraphSemiring(Semiring):
    def __init__(self, model, substitution, values):
        Semiring.__init__(self, model, substitution, values)

    def negate(self, a):
        return 1.0 - a

    def one(self):
        return 1.0

    def zero(self):
        return 0.0

    def plus(self, a, b):
        if self.is_zero(b):
            return a
        if self.is_zero(a):
            return b
        return a + b

    def times(self, a, b):
        if self.is_one(b):
            return a
        if self.is_one(a):
            return b
        return a * b

    def value(self, a, key=None):
        if type(a) is Constant:
            return float(a)
        elif type(a) is float:
            return a
        elif type(a) is Term:
            if a.functor == "nn":
                net, inputs = a.args[0], a.args[1]
                inputs = inputs.apply_term(self.substitution)
                val = self.values[(net, inputs)]
                i = 0
                if len(a.args) == 3:
                    i = int(a.args[2])
                return val[i]
            elif a.functor == "t":
                i = int(a.args[0])
                p = torch.tensor(self.model.parameters[i], requires_grad=True)
                p.register_hook(get_hook(self.model.optimizer, i))
                return p
            elif a.functor == "tensor":
                return self.model.get_tensor(a)
            elif a.functor == "'/'":  # Deals with probabilities formatted as franctions
                return float(a)
            else:
                raise Exception("unhandled term {}".format(a.functor))
        else:
            return float(a.compute_value())

    def is_one(self, a):
        return 1.0 - self.eps <= float(a) <= 1.0 + self.eps

    def is_zero(self, a):
        return -self.eps <= float(a) <= self.eps

    def is_dsp(self):
        return True

    def normalize(self, a, z):
        if self.is_one(z):
            return a
        return a / z

    @staticmethod
    def cross_entropy(
        result: Result,
        target: float,
        weight: float,
        q: Optional[Term] = None,
        eps: float = 1e-12,
    ) -> float:

        result = result.result
        if len(result) == 0:
            print("No results found for {}".format(q))
            return 0
        if q is None:
            if len(result) == 1:
                q, p = next(iter(result.items()))
            else:
                raise ValueError(
                    "q is None and number of results is {}".format(len(result))
                )
        else:
            p = result[q]
        if type(p) is float:
            loss = (
                -(target * math.log(p + eps) + (1.0 - target) * math.log(1.0 - p + eps))
                * weight
            )
        else:
            if target == 1.0:
                loss = -torch.log(p + eps) * weight
            elif target == 0.0:
                loss = -torch.log(1.0 - p + eps) * weight
            else:
                loss = (
                    -(
                        target * torch.log(p + eps)
                        + (1.0 - target) * torch.log(1.0 - p + eps)
                    )
                    * weight
                )
            loss.backward(retain_graph=True)
        return float(loss)

    @staticmethod
    def mse(
        result: Result, target: float, weight: float, q: Optional[Term] = None
    ) -> float:

        result = result.result
        if len(result) == 0:
            print("No results found for {}".format(q))
            return 0
        if q is None:
            if len(result) == 1:
                q, p = next(iter(result.items()))
            else:
                raise ValueError(
                    "q is None and number of results is {}".format(len(result))
                )
        else:
            p = result[q]
        loss = (p - target) ** 2 * weight
        if type(p) is not float:
            loss.backward(retain_graph=True)
        return float(loss)



from collections import defaultdict
from typing import TYPE_CHECKING

from torch import Tensor

if TYPE_CHECKING:
    from .model import Model


class Optimizer(object):
    """
    The optimizer responsible for optimizing the neural parameters in the model.
    """

    def __init__(self, model: "Model"):
        """
        :param model: The model whose parameters to optimize.
        """
        self.model = model
        self._params_grad = defaultdict(float)
        self.epoch = 0

    def step_epoch(self):
        """
        Perform an epoch step. This calls the step_epoch method of all the networks in the model.
        :return:
        """
        self.epoch += 1
        for _, network in self.model.networks.items():
            network.step_epoch()

    def add_parameter_gradient(self, k, grad: Tensor):
        """
        Accumulate the gradient for a probabilistic parameter.
        :param k: The key of the parameter.
        :param grad: The gradient to accumulate.
        :return:
        """
        pass

    def zero_grad(self):
        """
        Zeroes all gradients.
        :return:
        """
        self._params_grad.clear()
        for _, network in self.model.networks.items():
            network.zero_grad()
        if self.model.embeddings is not None:
            self.model.embeddings.zero_grad()

    def step(self):
        """
        Perform an optimization step for all parameters.
        :return:
        """
        for _, network in self.model.networks.items():
            network.step()
        if self.model.embeddings is not None:
            self.model.embeddings.step()


class SGD(Optimizer):
    """
    An optimizer that also optimizes the probabilistic parameters in the model using stochastic gradient descent.
    """

    def __init__(self, model: "Model", param_lr: float):
        """

        :param model: The model whose parameters will be optimized.
        :param param_lr: The learning rate for the parameters.
        """
        Optimizer.__init__(self, model)
        self.param_lr = param_lr

    def get_lr(self) -> float:
        """

        :return: The learning rate for the probabilistic parameters.
        """
        return self.param_lr

    def add_parameter_gradient(self, k, grad: Tensor):
        self._params_grad[k] += grad

    def step(self):
        Optimizer.step(self)
        for k in self._params_grad:
            self.model.parameters[k] -= float(self.get_lr() * self._params_grad[k])
            self.model.parameters[k] = max(min(self.model.parameters[k], 1.0), 0.0)
        for group in self.model.parameter_groups:
            p_sum = sum(self.model.parameters[x] for x in group)
            for param in group:
                self.model.parameters[param] /= p_sum



import pickle
from os import PathLike
from pathlib import Path
from typing import Callable, Generic, Hashable, TypeVar, Union

from deepproblog.utils import check_path

K = TypeVar("K")
T = TypeVar("T")


class Cache(Generic[K, T]):
    def __init__(
        self,
        func: Callable[[K], T],
        cache: bool,
        cache_root: Union[None, str, PathLike],
        key_func: Callable[[K], Hashable] = lambda x: x,
    ):
        """Create cache

        :param func: Function this cache evaluates
        :param cache: If true, do in memory caching.
        :param cache_root: If not None, cache to files at the provided path.
        :param key_func: Convert the key into a hashable object if needed
        """
        self.func = func
        self.key_func = key_func
        self.cache = cache
        if cache is True:
            print("Caching ACs")
        self.cache_file = cache_root is not None
        self.cache_dict = dict()
        self.first = func
        if self.cache_file:
            self.cache_root = Path(cache_root)
            self.first = self.get_from_file
        if self.cache:
            self.first = self.get_from_dict

    def __getitem__(self, item: K) -> T:
        return self.first(item)

    def invalidate(self):
        """Invalidate entire cache."""
        self.cache_dict.clear()
        if self.cache_file:
            for p in self.cache_root.iterdir():
                p.unlink()

    def get(self, item: K) -> T:
        return self.first(item)

    def get_from_dict(self, item: K) -> T:
        """Implements dict based cache."""
        cache_key = self.key_func(item)
        result = self.cache_dict.get(cache_key)
        if result is None:
            if self.cache_file:
                result = self.get_from_file(item, cache_key)
            else:
                result = self.func(item)
                result.from_cache = False
            self.cache_dict[cache_key] = result
        else:
            result.from_cache = True
        return result

    def get_from_file(self, item: K, cache_key=None) -> T:
        """Implements file based cache."""
        if cache_key is None:
            cache_key = self.key_func(item)
        filepath = self.cache_root / str(hash(cache_key))
        result = None
        if filepath.exists():
            with open(filepath, "rb") as f:
                (key, result) = pickle.load(f)
            if key == cache_key:
                return result
            else:
                # Hash collision! Handle this by overwriting the cache with the new query
                result = None
        if result is None:
            result = self.func(item)
            check_path(filepath)
            with open(filepath, "wb") as f:
                pickle.dump((cache_key, result), f)
        return result



import dataclasses
from typing import Callable, TypeVar, List

import pytest

from deepproblog.utils.cache import Cache

K = TypeVar("K")
T = TypeVar("T")


@dataclasses.dataclass
class _ResultType:
    value: str


@pytest.fixture(
    params=[
        {"cache_dict": False, "cache_file": False},
        {"cache_dict": True, "cache_file": False},
        {"cache_dict": False, "cache_file": True},
        {"cache_dict": True, "cache_file": True},
    ]
)
def cache_factory(request, tmpdir):
    kwargs = {
        "cache": request.param["cache_dict"],
        "cache_root": tmpdir.join("cache_dir") if request.param["cache_file"] else None,
    }

    def create_cache(func: Callable[[K], T], key_func=lambda x: x) -> Cache[K, T]:
        return Cache[K, T](func=func, key_func=key_func, **kwargs)

    # Certain tests are only valid if a cache is used.
    create_cache.no_cache = not (
        request.param["cache_dict"] or request.param["cache_file"]
    )
    return create_cache


def test_basic(cache_factory):
    cache = cache_factory(lambda x: _ResultType(str(x)))
    assert cache[32] == _ResultType("32")
    assert cache[1] == _ResultType("1")
    # This should retrieve it from cache
    assert cache[32] == _ResultType("32")


class _TerribleHash:
    """Test class to test hash collisions"""

    def __init__(self, v: int):
        self.v = v

    def __eq__(self, other):
        return isinstance(other, _TerribleHash) and other.v == self.v

    def __hash__(self):
        """Intentionally terrible hash to test hash collisions"""
        return 5 if self.v > 5 else 1


def test_hash_collision(cache_factory):
    cache = cache_factory(lambda x: _ResultType(str(x.v)))
    assert cache[_TerribleHash(1)] == _ResultType("1")
    assert cache[_TerribleHash(6)] == _ResultType("6")
    assert cache[_TerribleHash(2)] == _ResultType("2")


def test_invalidate(cache_factory):
    """Test that cache is properly invalidated on a call to invalidate()"""
    hash_breaker = 3
    cache = cache_factory(lambda x: _ResultType(str(x + hash_breaker)))
    assert cache[1] == _ResultType("4")
    hash_breaker = 2
    if not cache_factory.no_cache:
        # If caching is used, this should not yet return the new value.
        assert cache[1] == _ResultType("4")
    cache.invalidate()
    assert cache[1] == _ResultType("3")


@dataclasses.dataclass
class UnhashableData:
    cache_key: int
    unhashable_data: List[int]


def test_unhashable(cache_factory):
    if cache_factory.no_cache:
        pytest.skip("Unsupported when cache is not used.")
    cache = cache_factory(lambda x: _ResultType(str(x.cache_key)))
    with pytest.raises(TypeError):
        assert cache[UnhashableData(1, [])] == _ResultType("1")


def test_key_func(cache_factory):
    cache = cache_factory(
        lambda x: _ResultType(str(x.cache_key)), key_func=lambda x: x.cache_key
    )
    assert cache[UnhashableData(1, [])] == _ResultType("1")
    assert cache[UnhashableData(1, [1, 2, 3])] == _ResultType("1")



from pathlib import Path


if __name__ == '__main__':
    import sys
    if len(sys.argv) > 1:
        if sys.argv[1] == 'test':
            import pytest
            from pathlib import Path
            print(Path(__file__).parent)
            pytest.main([str(Path(__file__).parent)])



from math import isnan

import torch


class Heuristic(object):
    def __init__(self, name):
        self.name = name

    def get_hyperparameters(self):
        return {"type": type(self).__name__}

    def __repr__(self):
        return self.name


class PartialProbability(Heuristic):
    def __init__(self):
        super().__init__("partial_probability")


class GeometricMean(Heuristic):
    def __init__(self):
        super().__init__("geometric_mean")


class LearnedHeuristic(Heuristic):
    def __init__(self, name, predicates):
        super().__init__(name)
        self.predicates = predicates

    def count(self, batch, acs):
        raise NotImplementedError("Count method is not implemented.")

    def get_value(self, node):
        raise NotImplementedError

    def get_hyperparameters(self):
        return {"type": type(self).__name__, "predicates": self.predicates}


class NeuralHeuristic(LearnedHeuristic):
    def __init__(self, heuristic_functions, model):
        predicates = [x for x in heuristic_functions]
        name = "extern([{}])".format(",".join(p for p in predicates))
        super().__init__(name, predicates)
        self.heuristic_functions = heuristic_functions
        self.heuristic_optimizers = dict()
        self.cache = dict()
        self.model = model
        self.seen = set()
        self.t = 0
        self.freeze = False

    def count(self, batch, acs):
        if self.freeze:
            return None
        acs = list(zip(*acs))
        loss = torch.zeros(1)
        n = 0
        for i, (ac, semiring) in enumerate(acs):
            names = ac.get_named()
            for name in names:
                if name in self.cache:
                    target = 1.0
                    value = self.cache[name]
                    if target == 1.0:
                        newloss = -torch.log(value)
                    else:
                        newloss = -(
                            target * torch.log(value)
                            + (1.0 - target) * torch.log(1.0 - value)
                        )
                    if isnan(newloss):
                        raise ValueError("nan loss")
                    loss += newloss
                    n += 1
        if n > 0:
            loss /= n
            loss.backward(retain_graph=True)
        self.cache.clear()
        self.t += 1

    def get_value(self, node):

        value = self.heuristic_functions[node.functor](*node.args)
        self.cache[node] = value
        return float(value)

    def arg_to_tensor(self, arg):
        if arg.functor == "tensor":
            tensor = self.model.get_tensor(arg).detach()
            return tensor
        return self.model.get_embedding(arg).detach()
